!
! This is the intellectual property of Konrad Byers - Copyright 2002
! All rights reserved worldwide
!
! This template must not be copied, or modified, in part or whole, or distributed in any
! way, other than by the author Konrad Byers.
!
! No warranty of any kind is expressed, or implied. Used only at your own risk.
!
! Contact Konrad@eSoftAnywhere.com / ve1exe@rac.ca for additional information
!
!

!Any Acsii File Code

! These class methods are subject to a maximum ANYASCII_IO_SIZE line length per I/O request

 MEMBER

 MAP
    BOOLVAL(ANY AnyValue),BYTE
    StatusWith(ULONG FileOpenStatus, ULONG StatusToSet),ULONG
    MakeEvalStr(STRING StrToMakeEvalStr),STRING
    StrSlice(STRING StrToSlice, ULONG StartPosition=1, <ULONG StopPos>),STRING
    StrOfLength(STRING StrForSetLength, ULONG StrLength, <STRING FillChar>),STRING
    StructureItems(*GROUP GroupToQuery),ULONG
 END

 INCLUDE('Errors.clw'),ONCE
 INCLUDE('Keycodes.clw'),ONCE
 INCLUDE('AnyAscii.inc'),ONCE


AnyAsciiFileName        CSTRING(FILE:MaxFilePath+FILE:MaxFileName+1)
AnyAsciiFile            FILE,DRIVER('ASCII', '/FILEBUFFERS=64000'),CREATE,NAME(AnyAsciiFileName),PRE(AsciiFile)
                          RECORD
TextLine                    STRING(ANYASCII_IO_SIZE)
                          END
                        END


FieldNamesType  QUEUE,TYPE
RecordGroup         ANY
Field               ANY
FieldName           CSTRING(101),AUTO
FieldNumber         USHORT,AUTO
                END

!-----------------------------------------------------------------------------------------
!Construct
! Automatic constructor to initialize the object
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Construct        PROCEDURE
  CODE

  IF SELF.DefaultErrors &= NULL
     SELF.DefaultErrors &= NEW(ErrorClass)
     ASSERT(~SELF.DefaultErrors &= NULL)
     SELF.DefaultErrors.Init()
     SELF.Errors &= SELF.DefaultErrors
  END

  IF SELF.FieldNames &= NULL
     SELF.FieldNames &= NEW(FieldNamesType)
     ASSERT(~SELF.FieldNames &= NULL)
     FREE(SELF.FieldNames)
     CLEAR(SELF.FieldNames)
  END

  SELF.CtrlZ = 'ON'
  SELF.EndOfRecord = '<13><10>'
  SELF.Tab = CLIP(LEFT(8))


!-----------------------------------------------------------------------------------------
!Destruct
! Automatic destructor to ensure the file is closed properly
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Destruct          PROCEDURE
  CODE

  SELF.Close()

  IF ~SELF.DefaultErrors &= NULL
     SELF.DefaultErrors.Kill()
     DISPOSE(SELF.DefaultErrors)
     SELF.DefaultErrors &= NULL
  END

  IF ~SELF.FieldNames &= NULL
     FREE(SELF.FieldNames)
     CLEAR(SELF.FieldNames)
     DISPOSE(SELF.FieldNames)
     SELF.FieldNames &= NULL
  END


!-----------------------------------------------------------------------------------------
!Init
! Sets the error class to use (optional)
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Init              PROCEDURE(<ErrorClass EC>)
  CODE

  IF OMITTED(2)
     SELF.Errors &= EC
  ELSE
     SELF.Errors &= SELF.DefaultErrors
  END


!-----------------------------------------------------------------------------------------
!SetRecordLayout
! Initialize the class for use with Next, Insert, Update, Replace methods that do not
! receive a string type paramater but rather use an implied group structure specified
! with this method.
!
! * If FieldSeperator is omitted then a fixed length record format (determined by the
! RecordGroup GROUP) is used. If FieldSeperator is specified then a field FieldEnclosure
! may be specified. To use a fixed length record format exactly as specified by the
! RecordGroup GROUP paramater the call would be ...SetRecordLayout(SomeGroupName). To
! use a comma delimited record for the call would be
! ...SetRecordFormat(SomeGroupName, ',') and with quote enclosed fields it would become
! ...SetRecordFormat(SomeGroupName, ',', '"'). Tab character delimited records would be
! set as SetRecordFormat(SomeGroupName, CHR(TabKey), '"').
!
! * The SetTabs method is automatically called to enable correct TAB character interpretation
! if a TAB ('<4>') character is used for the field seperator character or the field
! enclosure character.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetRecordLayout   PROCEDURE(*GROUP RecordGroup, <STRING FieldSeperator>, <STRING FieldEnclosure>)
  CODE

  SELF.RecordGroup &= RecordGroup

  SELF.FieldSeperator = CHOOSE(BOOLVAL(OMITTED(3)), '', CLIP(SUB(FieldSeperator, 1, 1)))
  SELF.FieldEnclosure = CHOOSE(BOOLVAL(OMITTED(4)), '', CLIP(SUB(FieldEnclosure, 1, 1)))

  ASSERT(~SELF.FieldEnclosure OR SELF.FieldSeperator, |
        'Only seperated value files may utilize a "field enclosure character"')
  IF SELF.RecordGroup &= NULL OR (SELF.FieldEnclosure AND ~SELF.FieldSeperator)
     SELF.Errors.Throw(Msg:NoRecords)
     SELF.RecordGroup &= NULL
     SELF.FieldEnclosure = ''
     SELF.FieldSeperator = ''
     RETURN LEVEL:Notify
  END

  IF SELF.FieldSeperator = TabKey OR SELF.FieldEnclosure = TabKey
     SELF.SetTab()
  END

  SELF.PopulateRecord(False)

  RETURN LEVEL:Benign


!-----------------------------------------------------------------------------------------
!SetFieldName
! Sets a fields external name (the column heading name).
!
! When reading records, having set field names determines that the specified fields in the
! specified record/group structure will obtain their values from the columns of that name
! in the file (when the .SetProcessHeaderRecord behaviour is turned on to allow header record
! interpretation).
!
! When writing records, having set field names determines that the specified fields in the
! specified record/group structure will place/write their values in columns of that name
! as specified in the header record of the file (when the .SetWriteHeaderRecord behaviour
! is turned on to allow writing a header record).
!
! The .SetWriteHeaderRecord method determines whether a column header record (using these
! field names or the record/group field names by default) is written out as the first record.
! Otherwise there is really no need to set field names for writing records but READING
! records will still use these field names if specified to do so as described below using
! the SetProcessHeaderRecord method.
!
! The .SetProcessHeaderRecord method determines if the first record read should be processed
! as a header record and field mappings from column names in the file to record/group fields
! (as determined by the external field name specified using this (SetFieldName) method call)
! should be applied. Otherwise there is really no need to set field names for reading records
! but WRITING records will still use these field names if specified to do so as described above
! using the SetWriteHeaderRecord method.
!
! * Field names will otherwised be assumed from record/group field names when not specified
! specifically using this (SetFieldName) method call.
!
! ** See Also: SetFieldNumber
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetFieldName      PROCEDURE(*GROUP RecordGroup, *ANY Field, STRING FieldName)
 CODE

 SELF.SetField(RecordGroup, Field, SELF.FieldNames.FieldName, FieldName)


!-----------------------------------------------------------------------------------------
!SetFieldNumber
! Sets a fields external column number (the column number in the file).
!
! When reading records, having set field numbers determines that the specified field in the
! specified record/group structure will obtain its value from the nth (FieldNumber) field
! in the file.
!
! When writing records, having set field names determines that the specified fields in the
! specified record/group structure will place/write their values in columns of that name
! as specified in the header record of the file (when the .SetWriteHeaderRecord behaviour
! is turned on to allow writing a header record).
!
! Field numbers apply only when reading records (any generally only when no other reliable
! means of mapping file columns to record/group fields is possible.) When the file format
! is unknown so that simply calling SetRecordLayout will not suffice, and there is no header
! record available to provide names for the file columns that could be mapped by SetFieldName
! method calls, then SetFieldNumber calls are the only option left.
!
! * Field numbers are otherwised assumed from record/group field order by default when not
! specified specifically using this (SetFieldNumber) method call and when fields mappings
! are not otherwise established via SetFieldName method calls.
!
! ** See Also: SetFieldName
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetFieldNumber   PROCEDURE(*GROUP RecordGroup, *ANY Field, USHORT FieldNumber)
 CODE

 SELF.SetField(RecordGroup, Field, SELF.FieldNames.FieldNumber, FieldNumber)


!-----------------------------------------------------------------------------------------
!SetField
! Internal method to set field properties
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetField          PROCEDURE(*GROUP RecordGroup, *ANY Field, *ANY FieldProp, *ANY FieldValue)
Cntr                                ULONG,AUTO
FieldPresent                        BYTE(False)
 CODE

 ASSERT(~SELF.RecordGroup &= NULL, 'Record layout must be set')
 ASSERT(SELF.FieldSeperator, 'Record layout must be set')
 ASSERT(~SELF.FieldNames &= NULL)

 LOOP Cntr = 1 TO RECORDS(SELF.FieldNames)
    CLEAR(SELF.FieldNames)
    GET(SELF.FieldNames, Cntr)
    IF SELF.FieldNames.RecordGroup &= SELF.RecordGroup AND SELF.FieldNames.Field &= Field
       FieldProp = FieldValue
       PUT(SELF.FieldNames)
       ASSERT(~ERRORCODE())
       FieldPresent = True
       BREAK
    END
 END
 IF ~FieldPresent
    CLEAR(SELF.FieldNames)
    SELF.FieldNames.RecordGroup &= RecordGroup
    SELF.FieldNames.Field &= Field
    FieldProp = FieldValue
    ADD(SELF.FieldNames)
    ASSERT(~ERRORCODE())
 END


!-----------------------------------------------------------------------------------------
!ClearFields
! Resets field name and field order assignments previously specified with the SetFieldName
! and SetFieldNumber methods. If no record/group structure is specified then all such field
! properties are cleared otherise fields properties for fields of the specified record/group
! structure only are cleared.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.ClearFields       PROCEDURE(<*GROUP RecordGroup>)
Cntr                                ULONG,AUTO
RestartLoop                         BYTE,AUTO
  CODE

  IF OMITTED(2)
     FREE(SELF.FieldNames)
     CLEAR(SELF.FieldNames)
     RETURN
  END

  LOOP
     RestartLoop = False
     LOOP Cntr = 1 TO RECORDS(SELF.FieldNames)
        GET(SELF.FieldNames, Cntr)
        IF SELF.FieldNames.RecordGroup &= RecordGroup
           CLEAR(SELF.FieldNames)
           DELETE(SELF.FieldNames)
           RestartLoop = True
           BREAK
        END
     END
     IF ~RestartLoop THEN BREAK.
  END


!-----------------------------------------------------------------------------------------
!SetWriteHeaderRecord
! Specifies whether a column header record should be written as the first record of an
! ouput file
! * Individual column header / field names may be specified via .SetFieldName. Fields that
! have not specifically had their name set using the .SetFieldName method will assume their
! actual field names.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetWriteHeaderRecord PROCEDURE(BYTE WriteHeaderRecord)
 CODE

 ASSERT(~SELF.RecordGroup &= NULL, 'Record layout must be set')

 SELF.WriteHeaderRecord = WriteHeaderRecord


!-----------------------------------------------------------------------------------------
!SetProcessHeaderRecord
! Specifies whether the first record of an input file should be assumed to be a column
! header record and field mapping determinations made based on header record field names.
!
! * Individual column header / field name mappings may be specified via the .SetFieldName
! method. When this behaviour is enabled any fields that have not specifically had their
! name set using the .SetFieldName method will use their actual field names in the
! record/group structure so that all fields do actually have a field name (implied or
! otherwise).
!
! *** This behaviour is mutualy exclusive with SetIgnoreHeaderRecord behaviour. A header
! record can either be ignored or processed, one or the other but not both. Calling
! one turns the other off.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetProcessHeaderRecord PROCEDURE(BYTE ProcessHeaderRecord)
 CODE

 ASSERT(~SELF.RecordGroup &= NULL, 'Record layout must be set')

 SELF.ProcessHeaderRecord = ProcessHeaderRecord
 SELF.IgnoreHeaderRecord = False


!-----------------------------------------------------------------------------------------
!SetIgnoreHeaderRecord
! Sometimes a header record exists but it cannot be (or is not desirable to be) used for
! anything and so it must be ignored. Setting SetIgnoreHeaderRecord behaviour will cause
! the first record in the file to be automatically discarded.
!
! *** This behaviour is mutualy exclusive with SetProcessHeaderRecord behaviour. A header
! record can either be ignored or processed, one or the other but not both. Calling
! one turns the other off.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetIgnoreHeaderRecord PROCEDURE(BYTE IgnoreHeaderRecord)
 CODE

 ASSERT(~SELF.RecordGroup &= NULL, 'Record layout must be set')

 SELF.IgnoreHeaderRecord = IgnoreHeaderRecord
 SELF.ProcessHeaderRecord = False


!-----------------------------------------------------------------------------------------
!SetRecordFilter
! Sets a group/record filter for determining if each record should be processed.
! If the specified expression evaluates to true then the record is read otherwise the
! record is ommitted and the next record is read.
! ANYASCIIFILTER:BooleanExpression (indicating the string expression has not previously been
! formated for use as a filter or evaluate string) or ANYASCIIFILTER:FilterExpression
! (indicating it is a valid such type of string and need not be pre-processed to be so).
! As with any PROP:Filter statement all symbols in the evaluation must be 'BINDABLE'.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetRecordFilter   PROCEDURE(STRING RecordFilter, BYTE FilterType=ANYASCIIFILTER:BooleanExpression)
 CODE

 ASSERT(~SELF.RecordGroup &= NULL, 'Record layout must be set')

 CASE FilterType
 OF ANYASCIIFILTER:BooleanExpression
    SELF.RecordFilter = CLIP(MakeEvalStr(RecordFilter))
 OF ANYASCIIFILTER:FilterExpression
    SELF.RecordFilter = CLIP(RecordFilter)
 END


!-----------------------------------------------------------------------------------------
!ValidateRecord
! Internal virtual placeholder method
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.ValidateRecord    PROCEDURE()
 CODE

 RETURN Record:Ok


!-----------------------------------------------------------------------------------------
!PopulateRecord
! Internal method to move a line of retrieved text to a record layout group (as specified
! by the SetRecordLayout method)
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.PopulateRecord    PROCEDURE(BYTE HandleError)
Cntr                                ULONG(0)
Char                                STRING(1),AUTO
LookedAhead                         BYTE(False)
ParsedStr                           ANY
FieldNumber                         USHORT(0)
GroupField                          ANY
  CODE

  IF SELF.RecordGroup &= NULL OR (SELF.FieldEnclosure AND ~SELF.FieldSeperator)
     IF HandleError THEN SELF.Errors.Throw(Msg:NoRecords).
     RETURN LEVEL:Notify
  END

  IF ~SELF.FieldSeperator
     SELF.RecordGroup = StrOfLength(SELF.cTextLine, SIZE(SELF.RecordGroup))
     RETURN LEVEL:Benign
  END

  ParsedStr = ''
  LOOP

     IF ~LookedAhead
        Cntr += 1
        IF Cntr > LEN(CLIP(SELF.cTextLine)) THEN BREAK.
        Char = SELF.cTextLine[Cntr]
     ELSE
        LookedAhead = False
     END

     CASE Char
     OF SELF.FieldEnclosure ! Literals may contain spaces, keywords, etc if enclosed
        ParsedStr = ParsedStr & SELF.cTextLine[Cntr]
        LOOP
           Cntr += 1
           IF Cntr > LEN(CLIP(SELF.cTextLine)) THEN BREAK.
           Char = SELF.cTextLine[Cntr]
           IF Char = SELF.FieldEnclosure
              ParsedStr = ParsedStr & SELF.cTextLine[Cntr]
              Cntr += 1
              IF Cntr > LEN(CLIP(SELF.cTextLine)) THEN BREAK.
              Char = SELF.cTextLine[Cntr]
              IF Char ~= SELF.FieldEnclosure
                 LookedAhead = True
                 BREAK
              END
           ELSE
              ParsedStr = ParsedStr & SELF.cTextLine[Cntr]
           END
        END
     OF SELF.FieldSeperator
        FieldNumber += 1
        GroupField &= WHAT(SELF.RecordGroup, FieldNumber)
        IF GroupField &= NULL THEN BREAK.
        GroupField = ParsedStr
        GroupField &= NULL
        ParsedStr = ''
     ELSE
        ParsedStr = ParsedStr & SELF.cTextLine[Cntr]
     END

  END

  RETURN LEVEL:Benign


!-----------------------------------------------------------------------------------------
!UnpopulateRecord
! Internal method to move record layout group fields (as specified by the SetRecordLayout
! method) to a line of text
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.UnpopulateRecord  PROCEDURE(BYTE HandleError)
NumFields                           USHORT,AUTO
Cntr                                USHORT,AUTO
GroupField                          ANY
  CODE

  IF SELF.RecordGroup &= NULL OR ~SELF.FieldSeperator
     IF HandleError THEN SELF.Errors.Throw(Msg:NoRecords).
     RETURN LEVEL:Notify
  END

  SELF.cTextLine = ''
  NumFields = StructureItems(SELF.RecordGroup)
  LOOP Cntr = 1 TO NumFields
     GroupField &= WHAT(SELF.RecordGroup, Cntr)
     IF SELF.FieldEnclosure THEN SELF.cTextLine = SELF.cTextLine & SELF.FieldEnclosure.
     IF SELF.ClipFields
        SELF.cTextLine = SELF.cTextLine & CLIP(LEFT(GroupField))
     ELSE
        SELF.cTextLine = SELF.cTextLine & GroupField
     END
     IF SELF.FieldEnclosure THEN SELF.cTextLine = SELF.cTextLine & SELF.FieldEnclosure.
     IF Cntr < NumFields THEN SELF.cTextLine = SELF.cTextLine & SELF.FieldSeperator.
  END

  RETURN LEVEL:Benign


!-----------------------------------------------------------------------------------------
!ProcessHeaderFields
! Internal method to obtain record layout group field names from an input files header record
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.ProcessHeaderFields PROCEDURE()
  CODE

!-----------------------------------------------------------------------------------------
!TryOpen
! Initialize the class by opening an existing file to operate on
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryOpen           PROCEDURE(<STRING AsciiFileName>, <ULONG OpenMode>)
  CODE

  RETURN SELF.InitFile(CHOOSE(BOOLVAL(OMITTED(2)), SELF.FileName, AsciiFileName), |
                       SELF.Create, CHOOSE(BOOLVAL(OMITTED(3)), SELF.OpenMode, OpenMode), False)


!-----------------------------------------------------------------------------------------
!Open
! Initialize the class by opening an existing file to operate on
! * With error reporting
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Open              PROCEDURE(<STRING AsciiFileName>, <ULONG OpenMode>)
  CODE

  RETURN SELF.InitFile(CHOOSE(BOOLVAL(OMITTED(2)), SELF.FileName, AsciiFileName), |
                       SELF.Create, CHOOSE(BOOLVAL(OMITTED(3)), SELF.OpenMode, OpenMode), True)


!-----------------------------------------------------------------------------------------
!TryOpenExisting
! Initialize the class by creating a new (or truncating an existing) file to operate on
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryOpenExisting  PROCEDURE(<STRING AsciiFileName>, <ULONG OpenMode>)
  CODE

 RETURN SELF.InitFile(CHOOSE(BOOLVAL(OMITTED(2)), SELF.FileName, AsciiFileName), |
                      False, CHOOSE(BOOLVAL(OMITTED(3)), StatusWith(SELF.OpenMode, ReadOnly), |
                      OpenMode), False)


!-----------------------------------------------------------------------------------------
!OpenExisting
! Initialize the class by creating a new (or truncating an existing) file to operate on
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.OpenExisting     PROCEDURE(<STRING AsciiFileName>, <ULONG OpenMode>)
  CODE

  RETURN SELF.InitFile(CHOOSE(BOOLVAL(OMITTED(2)), SELF.FileName, AsciiFileName), |
                       False, CHOOSE(BOOLVAL(OMITTED(3)), StatusWith(SELF.OpenMode, ReadOnly),|
                       OpenMode), True)


!-----------------------------------------------------------------------------------------
!TryCreate
! Initialize the class by creating a new (or truncating an existing) file to operate on
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryCreateFile     PROCEDURE(<STRING AsciiFileName>, <ULONG OpenMode>)
  CODE

  RETURN SELF.InitFile(CHOOSE(BOOLVAL(OMITTED(2)), SELF.FileName, AsciiFileName), |
                       2, CHOOSE(BOOLVAL(OMITTED(3)), |
                       StatusWith(SELF.OpenMode, WriteOnly), |
                       OpenMode), False)


!-----------------------------------------------------------------------------------------
!CreateFile
! Initialize the class by creating a new (or truncating an existing) file to operate on
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.CreateFile        PROCEDURE(<STRING AsciiFileName>, <ULONG OpenMode>)
  CODE

  RETURN SELF.InitFile(CHOOSE(BOOLVAL(OMITTED(2)), SELF.FileName, AsciiFileName), |
                       2, CHOOSE(BOOLVAL(OMITTED(3)), |
                       StatusWith(SELF.OpenMode, WriteOnly), |
                       OpenMode), True)


!-----------------------------------------------------------------------------------------
!TryOpenOrCreate
! Initialize the class by opening an existing file or alternatively creating a new file
! when the file does not exist
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryOpenOrCreate   PROCEDURE(<STRING AsciiFileName>, <ULONG OpenMode>)
  CODE

  RETURN SELF.InitFile(CHOOSE(BOOLVAL(OMITTED(2)), SELF.FileName, AsciiFileName), |
                       True, CHOOSE(BOOLVAL(OMITTED(3)), |
                       StatusWith(SELF.OpenMode, WriteOnly), OpenMode), False)


!-----------------------------------------------------------------------------------------
!OpenOrCreate
! Initialize the class by opening an existing file or alternatively creating a new file
! when the file does not exist
! * With error reporting
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.OpenOrCreate      PROCEDURE(<STRING AsciiFileName>, <ULONG OpenMode>)
  CODE

  RETURN SELF.InitFile(CHOOSE(BOOLVAL(OMITTED(2)), SELF.FileName, AsciiFileName), |
                       True, CHOOSE(BOOLVAL(OMITTED(3)),  StatusWith(SELF.OpenMode, ReadWrite), |
                       OpenMode), True)


!-----------------------------------------------------------------------------------------
!TryNext
! Reads a line of text into the GROUP structure specified via the SetRecordFormat
! method. If no specific text file position is specified then the next file line in
! sequence is retrieved.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryNext           PROCEDURE(ULONG ReadPosition=0)
ReturnValue                         BYTE,AUTO
  CODE

  LOOP

     ReturnValue = SELF.NextInternal(ReadPosition, False)
     IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.

     ReturnValue = SELF.PopulateRecord(False)

     IF SELF.RecordFilter
        CASE EVALUATE(SELF.RecordFilter)
        OF ''
           ASSERT(~ERRORCODE(), ERROR())
           CYCLE
        OF '0'
           CYCLE
        END
     END
     IF SELF.ValidateRecord() ~= Record:Ok THEN CYCLE.

     BREAK

  END

  RETURN ReturnValue


!-----------------------------------------------------------------------------------------
!Next
! Reads a line of text into the GROUP structure specified via the SetRecordFormat
! method. If no specific text file position is specified then the next file line in
! sequence is retrieved.
! * With error reporting
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Next              PROCEDURE(ULONG ReadPosition=0)
ReturnValue                         BYTE,AUTO
  CODE

  LOOP

     ReturnValue = SELF.NextInternal(ReadPosition, True)
     IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.

     ReturnValue = SELF.PopulateRecord(True)

     IF SELF.RecordFilter
        CASE EVALUATE(SELF.RecordFilter)
        OF ''
           ASSERT(~ERRORCODE(), ERROR())
           CYCLE
        OF '0'
           CYCLE
        END
     END
     IF SELF.ValidateRecord() ~= Record:Ok THEN CYCLE.

     BREAK

  END

  RETURN ReturnValue


!-----------------------------------------------------------------------------------------
!TryNext
! Reads a line of text into the string buffer specified by TextLine
! If no specific text file position is specified then the next file line in sequence
! is retrieved.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryNext           PROCEDURE(*CSTRING TextLine, ULONG ReadPosition=0)
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.NextInternal(ReadPosition, False)
  TextLine = SELF.cTextLine
  RETURN ReturnValue

!-----------------------------------------------------------------------------------------
!Next
! Reads a line of text into the string buffer specified by TextLine
! If no specific text file position is specified then the next file line in sequence
! is retrieved.
! * With error reporting
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Next              PROCEDURE(*CSTRING TextLine, ULONG ReadPosition=0)
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.NextInternal(ReadPosition, True)
  TextLine = SELF.cTextLine
  RETURN ReturnValue


AnyAsciiFileClass.TryNext           PROCEDURE(*STRING TextLine, ULONG ReadPosition=0)
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.NextInternal(ReadPosition, False)
  TextLine = SELF.cTextLine
  RETURN ReturnValue

AnyAsciiFileClass.Next              PROCEDURE(*STRING TextLine, ULONG ReadPosition=0)
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.NextInternal(ReadPosition, True)
  TextLine = SELF.cTextLine
  RETURN ReturnValue


AnyAsciiFileClass.TryNext           PROCEDURE(*PSTRING TextLine, ULONG ReadPosition=0)
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.NextInternal(ReadPosition, False)
  TextLine = SELF.cTextLine
  RETURN ReturnValue

AnyAsciiFileClass.Next              PROCEDURE(*PSTRING TextLine, ULONG ReadPosition=0)
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.NextInternal(ReadPosition, True)
  TextLine = SELF.cTextLine
  RETURN ReturnValue


AnyAsciiFileClass.NextInternal      PROCEDURE(ULONG ReadPosition=0, BYTE HandleError)
ReturnValue                         BYTE(LEVEL:Benign)
ActualReadPosition                  ULONG,AUTO
 CODE

  ReturnValue = SELF.InitFile(,,, HandleError)
  IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.

  IF ~ReadPosition
     ActualReadPosition = SELF.ReadPosition
  ELSE
     ActualReadPosition = ReadPosition
  END

  LOOP

     GET(AnyAsciiFile, ActualReadPosition)
     IF ERRORCODE()
        SELF.SetError(Msg:AbortReading)
        RETURN CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
     END

     SELF.PreviousReadBytes = BYTES(AnyAsciiFile)
     SELF.PreviousReadPosition = ActualReadPosition
     SELF.ReadPosition = SELF.PreviousReadPosition + SELF.PreviousReadBytes

     IF ActualReadPosition ~= 1
        BREAK
     ELSIF SELF.ProcessHeaderRecord
        SELF.ProcessHeaderFields()
        ActualReadPosition = SELF.ReadPosition
        CYCLE
     ELSIF SELF.IgnoreHeaderRecord
        ActualReadPosition = SELF.ReadPosition
        CYCLE
     ELSE
        BREAK
     END


  END

  SELF.cTextLine = CLIP(AnyAsciiFile:TextLine)

  RETURN ReturnValue



!-----------------------------------------------------------------------------------------
!TryUpdate
! Replaces the previously read text with the text specified by the GROUP structure
! specified with the SetRecordLayout method
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryUpdate         PROCEDURE()
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.UnpopulateRecord(False)
  IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.
  RETURN SELF.UpdateInternal(False)


!-----------------------------------------------------------------------------------------
!Update
! Replaces the previously read text with the text specified by the GROUP structure
! specified with the SetRecordLayout method
! * With error reporting
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Update            PROCEDURE()
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.UnpopulateRecord(True)
  IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.
  RETURN SELF.UpdateInternal(True)



!-----------------------------------------------------------------------------------------
!TryUpdate
! Replaces the previously read line of text with the text specified by TextLine
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryUpdate         PROCEDURE(STRING TextLine)
  CODE

  SELF.cTextLine = TextLine
  RETURN SELF.UpdateInternal(False)


!-----------------------------------------------------------------------------------------
!Update
! Replaces the previously read line of text with the text specified by TextLine
! * With error reporting
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Update            PROCEDURE(STRING TextLine)
  CODE

  SELF.cTextLine = TextLine
  RETURN SELF.UpdateInternal(True)


!-----------------------------------------------------------------------------------------
!UpdateInternal
! Replaces the previously read line of text with the text specified by TextLine
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.UpdateInternal    PROCEDURE(BYTE HandleError)
PrevRec                             CSTRING(101),AUTO
PrevRecLen                          ULONG,AUTO
ReturnValue                         BYTE(LEVEL:Benign)
  CODE

  IF ~SELF.PreviousReadPosition
     SELF.SetError(Msg:PutFailed)
     RETURN CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
  END

  ReturnValue = SELF.InitFile(,,, HandleError)
  IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.

  PrevRec = SEND(AnyAsciiFile, 'ENDOFRECORD')
  PrevRecLen = SELF.PreviousReadBytes - LEN(PrevRec)

  GET(AnyAsciiFile, SELF.PreviousReadPosition)
  IF ERRORCODE()
     SELF.SetError(Msg:PutFailed)
     ReturnValue = CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
  ELSIF LEN(CLIP(SELF.cTextLine)) >= PrevRecLen
     AnyAsciiFile:TextLine = StrOfLength(SELF.cTextLine, PrevRecLen)
     PUT(AnyAsciiFile)
     IF ERRORCODE()
        SELF.SetError(Msg:PutFailed)
        ReturnValue = CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
     END
  ELSE
     ! A little magic going on here...
     ! We need to have a string with space padded ending. A dummy character is used to force
     ! a specific length string so that the driver will write the correct number of characters
     ! (and not just clip the string). Then because we really wrote a string that had a character
     ! we don't want there it is replaced with the space we really wished we could just put
     ! there in the first place. This is done by writing a blank string with the end of record
     ! character set to a space and to the necessary location in the file to undo the dummy
     ! character

     AnyAsciiFile:TextLine = StrOfLength(SELF.cTextLine, PrevRecLen - 1) & CHR(0)
     PUT(AnyAsciiFile)
     IF ERRORCODE()
        SELF.SetError(Msg:PutFailed)
        ReturnValue = CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
     END
     IF ReturnValue = LEVEL:Benign
        GET(AnyAsciiFile, SELF.PreviousReadPosition + PrevRecLen - 1)
        AnyAsciiFile:TextLine = ''
        SEND(AnyAsciiFile, 'ENDOFRECORD=1,32')
        PUT(AnyAsciiFile)
        IF ERRORCODE()
           SELF.SetError(Msg:PutFailed)
           ReturnValue = CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
        END
        SEND(AnyAsciiFile, 'ENDOFRECORD=' & PrevRec)
     END
  END

  IF SELF.AutoFlush THEN CLOSE(AnyAsciiFile).

  RETURN ReturnValue


!-----------------------------------------------------------------------------------------
!TryInsert
! Writes text contained in the GROUP structure specified with the SetRecordLayout method
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryInsert         PROCEDURE()
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.UnpopulateRecord(False)
  IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.
  RETURN SELF.InsertInternal(False)


!-----------------------------------------------------------------------------------------
!Insert
! Writes text contained in the GROUP structure specified with the SetRecordLayout method
! * With error reporting
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Insert            PROCEDURE()
ReturnValue                         BYTE,AUTO
  CODE

  ReturnValue = SELF.UnpopulateRecord(True)
  IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.
  RETURN SELF.InsertInternal(True)


!-----------------------------------------------------------------------------------------
!TryInsert
! Writes a line of text specified by TextLine to the end of the file
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.TryInsert         PROCEDURE(STRING TextLine)
  CODE

  SELF.cTextLine = TextLine
  RETURN SELF.InsertInternal(False)


!-----------------------------------------------------------------------------------------
!Insert
! Writes a line of text specified by TextLine to the end of the file
! * With error reporting
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Insert            PROCEDURE(STRING TextLine)
  CODE

  SELF.cTextLine = TextLine
  RETURN SELF.InsertInternal(True)


!-----------------------------------------------------------------------------------------
!Insert
! Writes a line of text specified by TextLine to the end of the file
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.InsertInternal    PROCEDURE(BYTE HandleError)
OrigTextLine                        ANY
NumFields                           USHORT,AUTO
Cntr                                USHORT,AUTO
FieldName                           ANY
Cntr2                               USHORT,AUTO
GroupField                          ANY
OrigGroupField                      ANY
ReturnValue                         BYTE(LEVEL:Benign)
  CODE

  ReturnValue = SELF.InitFile(,,, HandleError)
  IF ReturnValue ~= LEVEL:Benign THEN RETURN ReturnValue.

  IF SELF.WriteHeaderRecord AND ~SELF.WroteHeaderRecord
     OrigTextLine = SELF.cTextLine
     CLEAR(SELF.cTextLine)
     NumFields = StructureItems(SELF.RecordGroup)
     LOOP Cntr = 1 TO NumFields
        FieldName = ''
        LOOP Cntr2 = 1 TO RECORDS(SELF.FieldNames)
           GET(SELF.FieldNames, Cntr2)
           IF SELF.FieldNames.RecordGroup &= SELF.RecordGroup |
            AND SELF.FieldNames.FieldNumber = Cntr
              FieldName = SELF.FieldNames.FieldName
              BREAK
           END
        END
        IF ~CLIP(FieldName)
           FieldName = WHO(SELF.RecordGroup, Cntr)
           LOOP WHILE INSTRING(':', FieldName)
              FieldName = StrSlice(FieldName, INSTRING(':', FieldName) + 1)
           END
        END
        IF ~SELF.FieldSeperator
           GroupField &= WHAT(SELF.RecordGroup, Cntr)
           OrigGroupField = GroupField
           GroupField = FieldName
           FieldName = GroupField
           GroupField = OrigGroupField
           GroupField &= NULL
        END
        IF SELF.FieldEnclosure THEN SELF.cTextLine = SELF.cTextLine & SELF.FieldEnclosure.
        SELF.cTextLine = SELF.cTextLine & FieldName
        FieldName &= NULL
        IF SELF.FieldEnclosure THEN SELF.cTextLine = SELF.cTextLine & SELF.FieldEnclosure.
        IF Cntr < NumFields THEN SELF.cTextLine = SELF.cTextLine & SELF.FieldSeperator.
     END
  END

  FieldName      &= NULL
  GroupField     &= NULL
  OrigGroupField &= NULL

  IF SELF.WriteHeaderRecord AND ~SELF.WroteHeaderRecord
     AnyAsciiFile:TextLine = SELF.cTextLine
     ADD(AnyAsciiFile)
     SELF.cTextLine = OrigTextLine
     IF ERRORCODE()
        SELF.SetError(Msg:AddFailed)
        ReturnValue = CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
        IF SELF.AutoFlush THEN CLOSE(AnyAsciiFile).
        OrigTextLine &= NULL
        RETURN ReturnValue
     END
     SELF.WroteHeaderRecord = True
  END

  OrigTextLine &= NULL

  AnyAsciiFile:TextLine = SELF.cTextLine
  ADD(AnyAsciiFile)
  IF ERRORCODE()
     SELF.SetError(Msg:AddFailed)
     ReturnValue = CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
  END

  IF SELF.AutoFlush THEN CLOSE(AnyAsciiFile).

  RETURN ReturnValue


!-----------------------------------------------------------------------------------------
!Position
! Returns the current text file position in the read sequence
! (This is the position that will be read from by the read functions when no position
! is explicitly specified)
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Position          PROCEDURE()
  CODE

  IF ~SELF.FileOpened
     SELF.CheckFileOpened(False)
     RETURN 0
  ELSE
     RETURN SELF.ReadPosition
  END


!-----------------------------------------------------------------------------------------
!SetPosition
! Sets/Resets the current text file position in the read sequence
! (Sets the position that will be read from by the read functions when no position
! is explicitly specified)
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetPosition       PROCEDURE(ULONG ReadPosition)
  CODE

  IF SELF.FileOpened
     SELF.ReadPosition = ReadPosition
     RETURN LEVEL:Benign
  ELSE
     RETURN SELF.CheckFileOpened(False)
  END


!-----------------------------------------------------------------------------------------
!InsertPosition
! Returns the current text file position in the write sequence
! (This is the position that will be written to by the write functions, i.e. the number of
! bytes in the file + 1)
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.InsertPosition    PROCEDURE()
WritePosition                       ULONG(1)
  CODE

  IF ~SELF.FileOpened THEN RETURN 0.

  IF SELF.InitFile(,,, False) ~= LEVEL:Benign THEN RETURN 0.

  SET(AnyAsciiFile)
  LOOP
     NEXT(AnyAsciiFile)
     IF ERRORCODE() THEN BREAK.
     WritePosition = POINTER(AnyAsciiFile) + BYTES(AnyAsciiFile)
  END

  RETURN WritePosition


!-----------------------------------------------------------------------------------------
!SetName
! Used to set the default file name to assume when no file name is specified to the
! various methods that can open a file.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetName           PROCEDURE(STRING FileName)
  CODE

  SELF.FileName = FileName


!-----------------------------------------------------------------------------------------
!GetName
! Returns the current file name established by the SetName method only
! * See GetOpenName below
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.GetName           PROCEDURE()
  CODE

  RETURN SELF.FileName


!-----------------------------------------------------------------------------------------
!GetOpenName
! Returns the current/open file name of the file being operated on
! * This will not always be the same file name returned by the GetName method
! or specified with the SetName method as all the open/create file methods allow
! a file name override.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.GetOpenName       PROCEDURE()
  CODE

  RETURN SELF.CurrentFileName


!-----------------------------------------------------------------------------------------
!IsOpen
! Returns a boolean indication of whether or not the file is open
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.IsOpen            PROCEDURE()
  CODE

  RETURN SELF.FileOpened


!-----------------------------------------------------------------------------------------
!Close
! Closes the file being operated on
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.Close             PROCEDURE()
  CODE

  IF ~SELF.FileOpened THEN RETURN.
  CLOSE(AnyAsciiFile)
  SELF.FileOpened = False


!-----------------------------------------------------------------------------------------
!CheckFileOpened
! Internal method used to check if the file has been opened
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.CheckFileOpened  PROCEDURE(BYTE HandleError)
  CODE

  IF ~SELF.FileOpened
     SELF.SetError(Msg:UseClosedFile)
     RETURN CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
  ELSE
     RETURN LEVEL:Benign
  END


!-----------------------------------------------------------------------------------------
!SetAutoFlush
! Sets "Auto Flush" behaviour. When "Auto Flush" is enabled the file is closed after each
! disk write to force "flushing" the file. While this behaviour causes slower file I/O it
! may be desirable for logging and situates where a partially written file because of a
! program crash is desirable, etc.
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetAutoFlush      PROCEDURE(BYTE AutoFlush)
  CODE

  SELF.AutoFlush = AutoFlush


!-----------------------------------------------------------------------------------------
!SetCtrlZ
! Sets CTRL-Z = end of file handling on/off
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetCtrlZ          PROCEDURE(BYTE OnOff)
  CODE

  SELF.CtrlZ = CHOOSE(BOOLVAL(OnOff), 'ON', 'OFF')
  SEND(AnyAsciiFile, 'CTRLZISEOF=' & SELF.CtrlZ)
  SELF.CtrlZ = SEND(AnyAsciiFile, 'CTRLZISEOF=')


!-----------------------------------------------------------------------------------------
!SetEndOfRecord
! Sets the end of record indicator character sequence (the ENDOFRECORD ASCII driver setting)
! * By default this is '<13,10>' but mainframe file formats will often use '<13>'
! ** The end of record sequece is specified as a simple string of characters not in the
! more difficult ENDOFRECORD ASCII driver setting format
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetEndOfRecord    PROCEDURE(STRING EndOfRecord)
DriverEOR              ANY
Cntr                   USHORT(1)
  CODE

  SELF.EndOfRecord = ''
  LOOP Cntr = 1 TO LEN(EndOfRecord)
     SELF.EndOfRecord = CLIP(SELF.EndOfRecord) & CHOOSE(Cntr > 1, ',', '') |
                        & CLIP(LEFT(VAL(SUB(EndOfRecord, Cntr, 1))))
  END
  SEND(AnyAsciiFile, 'ENDOFRECORD=' & CHOOSE(LEN(EndOfRecord) > 0, |
                      CLIP(LEFT(LEN(EndOfRecord))) & ',' & SELF.EndOfRecord, '0'))
  SELF.EndOfRecord = SEND(AnyAsciiFile, 'ENDOFRECORD=')


!-----------------------------------------------------------------------------------------
!SetTabs
! Sends the TAB= ASCII driver command
! * The default (no Spaces parameter specified) turns off tab replacement. ** This behaviour
! is automatically set by the SetRecordLayout method if a TAB ('<4>') character is used
! for the field seperator character or the field enclosure character
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetTab            PROCEDURE(SHORT Spaces=-100)
  CODE

  SELF.Tab = CLIP(LEFT(Spaces))
  SEND(AnyAsciiFile, 'TAB=' & SELF.Tab)
  SELF.Tab = SEND(AnyAsciiFile, 'TAB=')


!-----------------------------------------------------------------------------------------
!SetClipFields
! Sets field value cliping behaviour when "writting records". The class can only "write
! records" when the .SetReclordLayout method has established a record layout. Clipping fixed
! length records or delimited value record can be performed to ensure all values written are
! string values even if the record layout contains non-string values (although using non-string
! values is generally not recommended). Additionally clipping delimited value records will
! change the values written to the file so that no white space appears in the value. For
! example a field value of "   some value   " will become "some value" (no leading or trailing
! spaces).
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.SetClipFields     PROCEDURE(BYTE ClipFields)
  CODE

  SELF.ClipFields = ClipFields


!-----------------------------------------------------------------------------------------
!FileStatusCurrent
! For internal use only (by .Open, .Close, etc.)
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.FileStatusCurrent PROCEDURE()
LastSelf          &AnyAsciiFileClass,STATIC
ReturnValue       BYTE(True)
  CODE

  IF ~LastSelf &= SELF
     LastSelf &= SELF
     RETURN False
     OMIT('***')
     RETURN CHOOSE(BOOLVAL(SELF.CurrentOpenMode) AND STATUS(AnyAsciiFile) = SELF.CurrentOpenMode |
                AND NAME(AnyAsciiFile) = SELF.CurrentFileName |
                AND UPPER(CLIP(SEND(AnyAsciiFile, 'CTRLZISEOF='))) = SELF.CtrlZ |
                AND CLIP(SEND(AnyAsciiFile, 'ENDOFRECORD=')) = SELF.EndOfRecord |
                AND CLIP(LEFT(SEND(AnyAsciiFile, 'TAB='))) = SELF.Tab)
     ***
  ELSE
     RETURN SELF.FileOpened
  END



!-----------------------------------------------------------------------------------------
!InitFile
! For internal use only (by .CreateFile & .Open, etc.)
!-----------------------------------------------------------------------------------------
AnyAsciiFileClass.InitFile          PROCEDURE(<STRING AsciiFileName>, BYTE CreateFile=False, LONG OpenMode=ReadWrite+DenyWrite, BYTE HandleError)
ReturnValue                         BYTE,AUTO
  CODE

  IF OMITTED(2)
     IF ~SELF.FileOpened THEN RETURN SELF.CheckFileOpened(HandleError).
     IF SELF.FileStatusCurrent() THEN RETURN LEVEL:Benign.
     CLOSE(AnyAsciiFile)
     AnyAsciiFileName = SELF.CurrentFileName
     SELF.CurrentOpenMode = CHOOSE(BOOLVAL(SELF.CurrentOpenMode), SELF.CurrentOpenMode, |
                                   ReadOnly+DenyNone)
  ELSE
     CLOSE(AnyAsciiFile)
     AnyAsciiFileName       = CLIP(AsciiFileName)
     SELF.CurrentFileName   = AnyAsciiFileName
     SELF.ReadPosition      = 1
     SELF.CurrentOpenMode   = CHOOSE(BOOLVAL(OpenMode), OpenMode, ReadOnly+DenyNone)
     SELF.FileOpened        = False
     SELF.WroteHeaderRecord = False
     CLEAR(SELF.LastError)
  END

  IF CreateFile = 2 OR (CreateFile AND ~EXISTS(AnyAsciiFileName))
     LOOP 50 TIMES
        REMOVE(AnyAsciiFileName)
        REMOVE(AnyAsciiFile)
        CREATE(AnyAsciiFile)
     WHILE ERRORCODE()
     IF ERRORCODE()
        SELF.SetError(Msg:CreateFailed)
        RETURN CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
     END
  END

  LOOP 50 TIMES
     OPEN(AnyAsciiFile, SELF.CurrentOpenMode)
     CASE ERRORCODE()
     OF NoError
        BREAK
     OF NoAccessErr
        IF ~OMITTED(2) AND HandleError AND BAND(SELF.CurrentOpenMode, ReadWrite)
           SELF.Throw(Msg:AccessDenied)
           OPEN(AnyAsciiFile, BAND(BXOR(SELF.CurrentOpenMode, ReadWrite), ReadOnly))
           IF ~ERRORCODE() THEN BREAK.
        END
     END
  END
  CASE ERRORCODE()
  OF NoError
  ELSE
     SELF.SetError(Msg:OpenFailed)
     RETURN CHOOSE(BOOLVAL(HandleError), SELF.Throw(), LEVEL:Notify)
  END

  SELF.CurrentOpenMode = STATUS(AnyAsciiFile)
  SELF.FileOpened      = True
  CLEAR(SELF.LastError)

  SEND(AnyAsciiFile, 'CTRLZISEOF=' & SELF.CtrlZ)
  SEND(AnyAsciiFile, 'ENDOFRECORD=' & SELF.EndOfRecord)
  SEND(AnyAsciiFile, 'TAB=' & SELF.Tab)

  RETURN LEVEL:Benign


AnyAsciiFileClass.SetError PROCEDURE(USHORT Err)
  CODE

  SELF.Errors.SetErrors()
  SELF.LastError = Err


AnyAsciiFileClass.GetError PROCEDURE
  CODE

  RETURN SELF.LastError


AnyAsciiFileClass.Throw PROCEDURE
  CODE

  SELF.Errors.SetFile(AnyAsciiFileName)
  RETURN SELF.Errors.TakeError(SELF.LastError)


AnyAsciiFileClass.Throw PROCEDURE(USHORT ErrNum)
  CODE

  SELF.SetError(ErrNum)
  RETURN SELF.Throw()


AnyAsciiFileClass.ThrowMessage PROCEDURE(USHORT ErrNum, STRING Me)
  CODE

  RETURN SELF.Errors.ThrowMessage(ErrNum, Me)




!-----------------------------------------------------------------------------!
!BOOLVAL
! Returns the boolean value of the input parameter
! * BOOLVAL(VariableName) is different from CHOOSE(VariableName) name, etc,...
! in that BOOLVAL forces a boolean logic only evaluation
!-----------------------------------------------------------------------------!
BOOLVAL                     PROCEDURE(ANY AnyValue)
  CODE

  IF AnyValue
     RETURN True
  ELSE
     RETURN False
  END


!-----------------------------------------------------------------------------!
!StatusWith
! Returns the result of a passed in STATUS return value or any valid OPEN
! access mode with the additional status mode specified by StatusToSet applied
! .i.e ReadOnly, WriteOnly returns ReadWrite
!-----------------------------------------------------------------------------!
StatusWith                 PROCEDURE(ULONG FileOpenStatus, ULONG StatusToSet)
OpenModeBits               ULONG,AUTO
HASREAD                    EQUATE(1)
HASWRITE                   EQUATE(2)
FileStatus                 ULONG(0)
SetStatus                  ULONG(0)
 CODE

 IF BAND(FileOpenStatus, ReadWrite) = ReadWrite OR BAND(FileOpenStatus, ReadOnly) = ReadOnly
    FileStatus += HASREAD;
 END
 IF BAND(FileOpenStatus, ReadWrite) = ReadWrite OR BAND(FileOpenStatus, WriteOnly) = WriteOnly
    FileStatus += HASWRITE;
 END

 IF BAND(StatusToSet, ReadWrite) = ReadWrite OR BAND(StatusToSet, ReadOnly) = ReadOnly
    SetStatus += HASREAD;
 END
 IF BAND(StatusToSet, ReadWrite) = ReadWrite OR BAND(StatusToSet, WriteOnly) = WriteOnly
    SetStatus += HASWRITE;
 END

 SetStatus = BOR(FileStatus, SetStatus)
 OpenModeBits = BOR(BOR(ReadWrite, WriteOnly), ReadOnly)
 StatusToSet = BXOR(BOR(StatusToSet, OpenModeBits), OpenModeBits)

 CASE SetStatus
 OF 0
 OF HASREAD
    StatusToSet += ReadOnly
 OF HASWRITE
    StatusToSet += WriteOnly
 OF HASREAD + HASWRITE
    StatusToSet += ReadWrite
 ELSE
    ASSERT(False)
 END

 RETURN BOR(FileOpenStatus, StatusToSet)


!---------------------------------------------------------------------------
! Returns an input string modified as necessary to be used in runtime evaluations
! (for the EVALUATE statement, PROP:Filter, etc.)
! * Doubles up < characters not enclosed in quotes
!---------------------------------------------------------------------------
MakeEvalStr         PROCEDURE(STRING StrToMakeEvalStr)
Cntr                ULONG(0)
Char                STRING(1),AUTO
LookedAhead         BYTE(False)
EvalStr             ANY
 CODE

 EvalStr = ''
 LOOP

    IF ~LookedAhead
       Cntr += 1
       IF Cntr > LEN(CLIP(StrToMakeEvalStr)) THEN BREAK.
       Char = StrToMakeEvalStr[Cntr]
    ELSE
       LookedAhead = False
    END

    CASE Char
    OF '<'
       EvalStr = EvalStr & '<<'
    OF '''' ! Literals may contain spaces, keywords, etc if quote enclosed
       EvalStr = EvalStr & StrToMakeEvalStr[Cntr]
       LOOP
          Cntr += 1
          IF Cntr > LEN(CLIP(StrToMakeEvalStr)) THEN BREAK.
          Char = StrToMakeEvalStr[Cntr]
          IF Char = ''''
             EvalStr = EvalStr & StrToMakeEvalStr[Cntr]
             Cntr += 1
             IF Cntr > LEN(CLIP(StrToMakeEvalStr)) THEN BREAK.
             Char = StrToMakeEvalStr[Cntr]
             IF Char ~= ''''
                LookedAhead = True
                BREAK
             END
          ELSE
             EvalStr = EvalStr & StrToMakeEvalStr[Cntr]
          END
       END
    ELSE
       EvalStr = EvalStr & StrToMakeEvalStr[Cntr]
    END

 END

 RETURN EvalStr


!---------------------------------------------------------------------------
! Returns a string that is the specified length by either truncating the input
! string or adding the necessary number of FillChar characters to the end of the
! string as required to create a string of the specified length. The FillChar is
! assumed to be ' ' i.e. SPACE when omitted.
!---------------------------------------------------------------------------
StrOfLength           PROCEDURE(STRING StrForSetLength, ULONG StrLength, <STRING FillChar>)
NewString             ANY
ActualFillChar        STRING(' ')
 CODE

 IF ~OMITTED(3) THEN ActualFillChar = SUB(FillChar, 1, 1).

 NewString = StrForSetLength & ALL(FillChar, StrLength - LEN(StrForSetLength))
 NewString = SUB(NewString, 1, StrLength)

 RETURN NewString


!---------------------------------------------------------------------------
! Returns a slice of a string from a start character position to a stop character position
! * The start character position defaults to the beginning of the string if ommitted
! * The stop character position defaults to then end of the string if ommitted
! i.e. ABCDEF, 2, 3 returns BCDE - ABCDEF,3 returns CDEF
!---------------------------------------------------------------------------
StrSlice            PROCEDURE(STRING StrToSlice, ULONG StartPosition=1, <ULONG StopPos>)
StopPosition        ULONG,AUTO
 CODE

 IF OMITTED(3)
    StopPosition = LEN(StrToSlice)
 ELSE
    StopPosition = StopPos
    IF StopPosition > LEN(StrToSlice) THEN StopPosition = LEN(StrToSlice).
 END

 IF ~StopPosition OR StartPosition > LEN(StrToSlice) OR StopPosition < StartPosition
    RETURN ''
 END

 StopPosition = StopPosition - StartPosition + 1

 RETURN SUB(StrToSlice, StartPosition, StopPosition)


!-----------------------------------------------------------------------------!
!StructureItems
! Returns the number of items/members in a specified structure
!-----------------------------------------------------------------------------!
StructureItems              PROCEDURE(*GROUP GroupToQuery)
Cntr                        USHORT(1)
GroupToQueryVar             ANY
  CODE

  LOOP
    GroupToQueryVar &= WHAT(GroupToQuery, Cntr)
    IF GroupToQueryVar &= NULL THEN BREAK.
    Cntr += 1
  END

  RETURN Cntr - 1



